import os
import importlib
import json

from queue import Empty
from urllib.parse import urljoin

import requests


def publish_batch(config, queue):
    result_list = []
    try:
        try:
            while not queue.empty() and len(result_list) < config.BUFFER_LENGTH:
                payload = queue.get_nowait()
                result_list.append(payload)
        except Empty:
            pass

        if len(result_list) == 0:
            return

        version = importlib.import_module(__package__).__version__
        url = urljoin(os.getenv("METRICS_SERVER", config.METRICS_API), "/v1/request")

        readme_result = requests.post(
            url,
            auth=(config.README_API_KEY, ""),
            data=json.dumps(result_list),
            headers={
                "Content-Type": "application/json",
                "User-Agent": f"readme-metrics-python@{version}",
            },
            timeout=config.METRICS_API_TIMEOUT,
        )
        config.LOGGER.info(
            f"POST to {url} with {len(result_list)} items returned {readme_result.status_code}"
        )
        if not readme_result.ok:
            config.LOGGER.exception(readme_result.text)
            raise Exception(f"POST to {url} returned {readme_result.status_code}")
    except Exception as e:
        # Errors in the Metrics SDK should never cause the application to
        # throw an error. Log it but don't re-raise.
        config.LOGGER.exception(e)
    finally:
        for _ in result_list:
            queue.task_done()



# pylint: disable=too-many-instance-attributes
import importlib
from typing import List, Any, Callable

from readme_metrics.util import util_build_logger


class MetricsApiConfig:
    """ReadMe Metrics API configuration object

    Attributes:
        README_API_KEY (str) Your ReadMe API key
        GROUPING_FUNCTION (lambda): Grouping function to construct an
            identity object. It receives the current request as a parameter, and must
            return a dictionary containing at least an "id" field, and optionally
            "label" and "email" fields.

            The main purpose of the identity object is to identify the API's caller.
        BUFFER_LENGTH (int): Number of requests to buffer before sending data
            to ReadMe. Defaults to 10.
        IS_DEVELOPMENT_MODE (bool): Determines whether you are running in
            development mode. Defaults to False.
        IS_BACKGROUND_MODE (bool):  Determines whether to issue the call to
            the ReadMe API in a background thread. Defaults to True.
        DENYLIST (List[str]): An array of headers and JSON body properties to
            skip sending to ReadMe.

            If you configure a denylist, it will override any allowlist configuration.
        ALLOWLIST (List[str]): An array of headers and JSON body properties to
            send to ReadMe.

            If this option is configured, ONLY the allowlisted properties will be sent.
        ALLOWED_HTTP_HOSTS (List[str]): A list of allowed http hosts for sending
            data to the ReadMe API.
        METRICS_API (str): Base URL of the ReadMe metrics API.
        METRICS_API_TIMEOUT (int): Timeout (in seconds) for metrics API calls.
        LOGGER (logging.Logger): Logger used by all classes and methods in the
            readme_metrics packge. Defaults to a basic console logger with log level
            CRITICAL.

            You can adjust logging settings by manipulating LOGGER, or you can replace
            LOGGER entirely with your application's Logger.
    """

    README_API_KEY: str = None
    BUFFER_LENGTH: int = 10
    GROUPING_FUNCTION: Callable[[Any], None] = lambda req: None
    IS_DEVELOPMENT_MODE: bool = False
    IS_BACKGROUND_MODE: bool = True
    DENYLIST: List[str] = []
    ALLOWLIST: List[str] = []
    ALLOWED_HTTP_HOSTS: List[str] = []
    METRICS_API: str = "https://metrics.readme.io"
    METRICS_API_TIMEOUT: int = 3

    def __init__(
        self,
        api_key: str,
        grouping_function,
        buffer_length: int = 10,
        development_mode: bool = False,
        background_worker_mode: bool = True,
        allowlist: List[str] = None,
        denylist: List[str] = None,
        blacklist: List[str] = None,
        whitelist: List[str] = None,
        allowed_http_hosts: List[str] = None,
        timeout: int = 3,
    ):
        """Initializes an instance of the MetricsApiConfig object

        Args:
            api_key (str): Your ReadMe API key
            grouping_function ([type] or str]): Grouping function to construct an identity
                object. It receives the current request as a parameter, and must return
                a dictionary containing at least an "id" field, and optionally "label"
                and "email" fields.

                The main purpose of the identity object is to identify the API's caller.

                You can optionally pass the path of the function to the MetricsApiConfig
                constructor, in which case it will automatically be resolved and imported
                when this object is initialized.
            buffer_length (int, optional): Number of requests to buffer before sending
                data to ReadMe. Defaults to 10.
            development_mode (bool, optional): Determines whether you are running in
                development mode. Defaults to False.
            background_worker_mode (bool, optional): Determines whether to issue the
                call to the ReadMe API in a background thread. Defaults to True.
            denylist (List[str], optional): An array of keys from your API requests and
                responses headers and bodies that are blocked from being sent to ReadMe.
                Both the request and response will be checked for these keys, in their
                HTTP headers, form fields, URL parameters, and JSON request/response
                bodies. JSON is only checked at the top level, so a nested field will
                still be sent even if its key matches one of the keys in `denylist`.
                Defaults to None.

                If you configure a denylist, it will override any allowlist
                configuration.
            allowlist (List[str], optional): An array of headers and JSON body
                properties to send to ReadMe. Similar semantics to `denylist`; defaults
                to None.

                If this option is configured, ONLY the whitelisted properties will be
                sent.
            blacklist (List[str], optional): Deprecated, prefer denylist.
            whitelist (List[str], optional): Deprecated, prefer allowlist.
            allowed_http_hosts (List[str], optional): A list of HTTP hosts which should be
                logged to ReadMe. If this is present, requests will only be sent to ReadMe
                whose Host header matches one of the allowed hosts.
            timeout (int): Timeout (in seconds) for calls back to the ReadMe Metrics API.
                Default 3 seconds.
        """
        self.README_API_KEY = api_key
        if isinstance(grouping_function, str):
            module_name, function_name = grouping_function.rsplit(".", 1)
            module = importlib.import_module(module_name)
            self.GROUPING_FUNCTION = getattr(module, function_name)
        else:
            self.GROUPING_FUNCTION = grouping_function
        self.BUFFER_LENGTH = buffer_length
        self.IS_DEVELOPMENT_MODE = development_mode
        self.IS_BACKGROUND_MODE = background_worker_mode
        self.DENYLIST = denylist or blacklist or []
        self.ALLOWLIST = allowlist or whitelist or []
        self.ALLOWED_HTTP_HOSTS = allowed_http_hosts
        self.METRICS_API_TIMEOUT = timeout
        self.LOGGER = util_build_logger()



import atexit
import math
import queue
import threading

from readme_metrics import MetricsApiConfig
from readme_metrics.publisher import publish_batch
from readme_metrics.PayloadBuilder import PayloadBuilder
from readme_metrics.ResponseInfoWrapper import ResponseInfoWrapper


class Metrics:
    """
    This is the internal central controller class invoked by the ReadMe middleware. It queues
    requests for submission for processing by `readme_metrics.publisher.publish_batch()`.
    """

    PACKAGE_NAME: str = "readme/metrics"

    def __init__(self, config: MetricsApiConfig):
        """
        Constructs and initializes the ReadMe Metrics controller class with the
        specified configuration.

        Args:
            config (MetricsApiConfig): Running configuration
        """

        self.config = config
        self.payload_builder = PayloadBuilder(
            config.DENYLIST,
            config.ALLOWLIST,
            config.IS_DEVELOPMENT_MODE,
            config.GROUPING_FUNCTION,
            config.LOGGER,
        )
        self.queue = queue.Queue()

        atexit.register(self.exit_handler)

    def process(self, request, response: ResponseInfoWrapper) -> None:
        """Enqueues a request/response combination to be submitted the API.

        Args:
            request (Request): Request object from your WSGI server
            response (ResponseInfoWrapper): Response object
        """
        if not self.host_allowed(request.environ["HTTP_HOST"]):
            # pylint: disable=C0301
            self.config.LOGGER.debug(
                f"Not enqueueing request, host {request.environ['HTTP_HOST']} not in ALLOWED_HTTP_HOSTS"
            )
            return

        payload = self.payload_builder(request, response)
        if payload is None:
            # PayloadBuilder returns None when the grouping function returns
            # None (an indication that the request should not be logged.)
            self.config.LOGGER.debug(
                "Not enqueueing request, grouping function returned None"
            )
            return

        self.queue.put(payload)
        if self.queue.qsize() >= self.config.BUFFER_LENGTH:
            args = (self.config, self.queue)
            if self.config.IS_BACKGROUND_MODE:
                thread = threading.Thread(target=publish_batch, daemon=True, args=args)
                thread.start()
            else:
                publish_batch(*args)

    def exit_handler(self) -> None:
        if not self.queue.empty():
            args = (self.config, self.queue)
            for _ in range(math.ceil(self.queue.qsize() / self.config.BUFFER_LENGTH)):
                if self.config.IS_BACKGROUND_MODE:
                    thread = threading.Thread(
                        target=publish_batch, daemon=True, args=args
                    )
                    thread.start()
                else:
                    publish_batch(*args)
        self.queue.join()

    def host_allowed(self, host):
        if self.config.ALLOWED_HTTP_HOSTS:
            return host in self.config.ALLOWED_HTTP_HOSTS

        # If `allowed_http_hosts`` has not been set (None by default), send off the data to be
        # queued.
        return True



import logging
from logging import Formatter, StreamHandler


def util_build_logger():
    logger = logging.getLogger(__package__)
    logger.setLevel(logging.CRITICAL)
    formatter = Formatter(fmt="%(asctime)s [%(levelname)s] %(message)s")
    handler = StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



# pylint: disable=too-many-locals
import io
import time
import datetime

from werkzeug import Request

from readme_metrics.Metrics import Metrics
from readme_metrics.MetricsApiConfig import MetricsApiConfig
from readme_metrics.ResponseInfoWrapper import ResponseInfoWrapper


class MetricsMiddleware:
    """Core middleware class for ReadMe Metrics

    Attributes:
        config (MetricsApiConfig): Contains the configuration settings for the running
            middleware instance
    """

    def __init__(self, wsgi_app_reference, config: MetricsApiConfig):
        """
        Constructs and initializes MetricsMiddleware WSGI middleware to be passed into
        the currently running WSGI web server.

        Args:
            wsgi_app_reference ([type]): Reference to the current WSGI application,
                which will be wrapped
            config (MetricsApiConfig): Instance of MetricsApiConfig object
        """
        self.config = config
        self.app = wsgi_app_reference
        self.metrics_core = Metrics(config)

    def __call__(self, environ, start_response):
        """Method that is called by the running WSGI server.

        You should NOT be calling this method yourself under normal circumstances.
        """
        response_headers = {}
        response_status = 0
        iterable = None
        req = Request(environ)

        def _start_response(_status, _response_headers, *args):
            write = start_response(_status, _response_headers, *args)

            # Populate response info (headers & status)
            nonlocal response_headers, response_status

            response_headers = _response_headers
            response_status = _status

            return write

        try:
            req.rm_start_dt = str(datetime.datetime.utcnow())
            req.rm_start_ts = int(time.time() * 1000)

            if req.method == "POST":
                # The next 4 lines are a workaround for a serious shortcoming in the
                # WSGI spec.
                #
                # The data can only be read once, after which the socket is exhausted
                # and cannot be read again. As such, we read the data and then
                # repopulate the variable so that it can be used by other code down the
                # pipeline.
                #
                # For more info: https://stackoverflow.com/a/13106009/643951

                # the environment variable CONTENT_LENGTH may be empty or missing
                try:
                    content_length = int(environ.get("CONTENT_LENGTH", 0))
                except ValueError:
                    content_length = 0
                content_body = environ["wsgi.input"].read(content_length)

                # guarding check to close stream
                if hasattr(environ["CONTENT_LENGTH"], "close"):
                    environ["wsgi.input"].close()

                environ["wsgi.input"] = io.BytesIO(content_body)

                req.rm_content_length = content_length
                req.rm_body = content_body

            iterable = self.app(environ, _start_response)
            for data in iterable:
                res_ctype = ""
                res_clength = 0

                htype = next(
                    (h for h in response_headers if h[0] == "Content-Type"), None
                )

                hlength = next(
                    (h for h in response_headers if h[0] == "Content-Length"), None
                )

                if htype and hlength:
                    res_ctype = htype[1]
                    res_clength = int(hlength[1])

                # Populate response body
                res = ResponseInfoWrapper(
                    response_headers,
                    response_status,
                    res_ctype,
                    res_clength,
                    data.decode("utf-8"),
                )

                # Send off data to be queued (and processed) by ReadMe if allowed
                self.metrics_core.process(req, res)

                yield data

        finally:
            # Undocumented in WSGI spec but the iterable has to be closed
            if hasattr(iterable, "close"):
                iterable.close()



import json
import hmac
from datetime import datetime, timedelta


class VerifyWebhook:
    def __init__(self, body, signature, secret):
        if signature is None:
            raise Exception("Missing Signature")

        parsed_input = dict(
            (x.strip(), y.strip())
            for x, y in (element.split("=") for element in signature.split(","))
        )

        time = parsed_input["t"]

        if datetime.now() - datetime.fromtimestamp(
            int(parsed_input["t"]) / 1000
        ) > timedelta(minutes=30):
            raise Exception("Expired Signature")

        unsigned = time + "." + json.dumps(body, separators=(",", ":"))
        verify_signature = hmac.new(
            secret,
            unsigned.encode("utf8"),
            "sha256",
        ).hexdigest()
        readme_signature = parsed_input["v0"]
        if (
            hmac.compare_digest(
                verify_signature.encode(), readme_signature.encode("utf8")
            )
            is False
        ):
            raise Exception("Invalid Signature")



import importlib
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

version = importlib.import_module("readme_metrics").__version__

setup(
    name="readme-metrics",
    version=version,
    author="ReadMe",
    author_email="support@readme.io",
    description="ReadMe API Metrics SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/readmeio/metrics-sdks/tree/main/packages/python",
    packages=["readme_metrics"],
    install_requires=["requests", "Werkzeug"],
    extras_require={"Flask": ["Flask"], "Django": ["Django"]},
)
